"""
Given some extended data (for example features), this is the class that handles it.
"""

import json
import os
import pandas as pd
from copy import deepcopy

UNKNOWN_SYM = "__UNKNOWN__"


class ExternalData:
    def __init__(self, params, idx_to_symbol=None):
        # load the params file (json) in the "external" section.
        self._params = params["external"] if type(params) is dict else json.load(open(params, "rt"))["external"]
        # path to base directory of the project
        self._base_dir = __file__.replace("/", os.sep)
        self._base_dir = os.path.join(self._base_dir.rsplit(os.sep, 1)[0], "../../..")
        # path to external data file

        # { ... graph_id: { node: [..value..] } ... }
        self._embed, self._continuous, self._idx_to_symbol, self._symbol_to_idx = self._read_file()

        # fix idx to sym and opposite
        self._idx_to_symbol = self._idx_to_symbol if idx_to_symbol is None else idx_to_symbol
        self._symbol_to_idx = self._symbol_to_idx if idx_to_symbol is None else  \
            {ftr: {symbol: i for i, symbol in enumerate(idx_to_symbol_list)}
             for ftr, idx_to_symbol_list in idx_to_symbol.items()}

    @property
    def idx_to_symbol_dict(self):
        return self._idx_to_symbol

    @property
    def is_embed(self):
        return False if self._embed is None else True

    @property
    def is_continuous(self):
        return False if self._continuous is None else True

    @property
    def embed_headers(self):
        return deepcopy(self._params["embeddings"])

    @property
    def continuous_headers(self):
        return deepcopy(self._params["continuous"])

    def embed_feature(self, g_id, node):
        # self._embed[g_id][node] = symbol_list of the node in graph
        return [self._symbol_to_idx[ftr].get(self._embed[g_id][node][i], self._symbol_to_idx[ftr][UNKNOWN_SYM])
                for i, ftr in enumerate(self._params["embeddings"])]

    # return number of symbols for specific embedding
    def len_embed(self, idx_str=None):
        if idx_str is None:
            lengths = []
            for ftr, symbol_list in self._idx_to_symbol.items():
                lengths.append(len(symbol_list))
            return lengths
        ftr = self._params["embeddings"][idx_str] if type(idx_str) == int else idx_str
        return len(self._idx_to_symbol[ftr])

    def continuous_feature(self, g_id, node):
        return self._continuous[g_id][node]

    # read file into two dictionaries one for embedding values and one for rational values.
    def _read_file(self):
        idx_to_symbol = {ftr: [UNKNOWN_SYM] for ftr in self._params["embeddings"]}

        embed_dict = {} if self._params["embeddings"] else None
        value_dict = {} if self._params["continuous"] else None

        external_data_df = pd.read_csv(self._params["file_path"])
        # read all rows
        for index, data in external_data_df.iterrows():
            # extract data
            g_id = str(data[self._params["graph_col"]])
            node = str(data[self._params["node_col"]])

            # extract embeddings + keep list idx to embed symbol for each feature
            embed_list = []
            for ftr in self._params["embeddings"]:
                symbol = str(data[ftr])
                embed_list.append(symbol)
                if symbol not in idx_to_symbol[ftr]:
                    idx_to_symbol[ftr].append(symbol)

            value_list = [data[ftr] for ftr in self._params["continuous"]]

            # add to dictionaries { ... graph_id: { node: [..list_value..] } ... }
            if embed_dict is not None:
                embed_dict[g_id] = embed_dict.get(g_id, {})
                embed_dict[g_id][node] = embed_list
            if value_dict is not None:
                value_dict[g_id] = value_dict.get(g_id, {})
                value_dict[g_id][node] = value_list

        symbol_to_idx = {ftr: {symbol: i for i, symbol in enumerate(idx_to_symbol_list)} for ftr, idx_to_symbol_list in
                         idx_to_symbol.items()}
        return embed_dict, value_dict, idx_to_symbol, symbol_to_idx



import json
import pickle
from itertools import combinations, permutations

import networkx as nx


class IsomorphismGenerator:
    def __init__(self, group_size, is_directed):
        self._group_size = group_size
        self._is_directed = is_directed
        graphs = self._generate_all_graphs()
        self._isomorphisms = self._group_to_isomorphisms(graphs)
        self._remove_irrelevant()
        self._reorganize()

    # Generate all possible graphs of size 'group_size'
    def _generate_all_graphs(self):
        # handshake lemma
        num_edges = int(self._group_size * (self._group_size - 1) / 2.)
        num_bits = num_edges * 2 if self._is_directed else num_edges
        edge_iter = permutations if self._is_directed else combinations
        graph_type = nx.DiGraph if self._is_directed else nx.Graph

        graphs = {}
        for num in range(2 ** num_bits):
            g = graph_type()
            g.add_nodes_from(range(self._group_size))
            g.add_edges_from((x, y) for i, (x, y) in enumerate(edge_iter(range(self._group_size), 2)) if (2 ** i) & num)
            graphs[num] = g
        return graphs

    @staticmethod
    def _group_to_isomorphisms(graphs):
        isomorphisms = {}
        i = 0
        keys = sorted(list(graphs.keys()))
        while keys:
            g1 = graphs[keys[0]]
            isomorphisms[i] = {num: graphs[num] for num in keys if nx.is_isomorphic(g1, graphs[num])}
            keys = [x for x in keys if x not in isomorphisms[i]]
            i += 1
        return isomorphisms

    def _remove_irrelevant(self):
        isomorphisms = self._isomorphisms
        # Remove disconnected graphs
        irrelevant = [n for n, gs in isomorphisms.items() if not nx.is_connected(list(gs.values())[0].to_undirected())]
        if irrelevant:
            isomorphisms[None] = {}
        for n in irrelevant:
            isomorphisms[None].update(isomorphisms.pop(n))

    def _reorganize(self):
        keys = [x for x in self._isomorphisms if x is not None]
        irrelevant = self._isomorphisms.get(None)
        self._isomorphisms = {i: self._isomorphisms[key] for i, key in enumerate(keys)}
        if irrelevant:
            self._isomorphisms[None] = irrelevant

    def num_2_motif(self):
        return {num: motif_num for motif_num, group in self._isomorphisms.items() for num in group}


def main(level, is_directed):
    fname = "%d_%sdirected" % (level, "" if is_directed else "un")
    print("Calculating ", fname)
    gs = IsomorphismGenerator(level, is_directed)
    # Json dump integers to strings (JavaScript compatibility), other option - to dump
    json.dump(list(gs.num_2_motif().items()), open(fname + ".json", "w"))
    pickle.dump(gs.num_2_motif(), open(fname + ".pkl", "wb"))
    print("Finished calculating ", fname)
    # for y in gs.values():
    #     print(list(map(lambda i: len(i.edges()), y.values())))


if __name__ == "__main__":
    main(3, False)
    main(3, True)
    main(4, False)
    main(4, True)
    # main(5, False)
    # main(5, True)
    print("Bla")



import pickle
from graphMeasures.features_infra.graph_features import GraphFeatures
import numpy as np
from graphMeasures.loggers import BaseLogger, PrintLogger
import os
MOTIFS_VAR_PATH = os.path.join(__file__.rsplit(os.sep, 1)[0])


class MotifRatio:
    def __init__(self, ftr: GraphFeatures, is_directed, logger: BaseLogger=None):
        self._is_directed = is_directed                 # are the graphs directed
        self._index_ftr = None                          # list of ftr names + counter [ ... (ftr_i, 0), (ftr_i, 1) ...]
        self._logger = logger if logger else PrintLogger("graphs logger")
        # self._graph_order = graph_order if graph_order else [g for g in sorted(graph_ftr_dict)]
        self._gnx_ftr = ftr
        self._set_index_to_ftr(self._gnx_ftr)
        # list index in motif to number of edges in the motif
        self._motif_index_to_edge_num = {"motif3": self._motif_num_to_number_of_edges(3),
                                         "motif4": self._motif_num_to_number_of_edges(4)}
        self._ftr_mx = self._gnx_ftr.to_matrix(dtype=np.float32, mtype=np.matrix, should_zscore=False)
        self._headers = []
        self._motif_ratio_vec = None
        self._motif_ratio_matrix = None

    # load motif variation file
    def _load_variations_file(self, level):
        fname = "%d_%sdirected.pkl" % (level, "" if self._is_directed else "un")
        fpath = os.path.join(MOTIFS_VAR_PATH, "motif_variations", fname)
        return pickle.load(open(fpath, "rb"))

    # return dictionary { motif_index: number_of_edges }
    def _motif_num_to_number_of_edges(self, level):
        motif_edge_num_dict = {}
        for bit_sec, motif_num in self._load_variations_file(level).items():
            motif_edge_num_dict[motif_num] = bin(bit_sec).count('1')
        return motif_edge_num_dict

    # map matrix columns to features + count if there's more then one from a single feature
    def _set_index_to_ftr(self, gnx_ftr):
        if not self._index_ftr:
            sorted_ftr = [f for f in sorted(gnx_ftr) if gnx_ftr[f].is_relevant()]  # fix feature order (names)
            self._index_ftr = []

            for ftr in sorted_ftr:
                len_ftr = len(gnx_ftr[ftr])
                # fill list with (ftr, counter)
                self._index_ftr += [(ftr, i) for i in range(len_ftr)]

    # get feature vector for a graph
    def _build_vector(self):
        # get gnx gnx
        final_vec = np.zeros((1, self._ftr_mx.shape[1]))

        motif3_ratio = None
        motif4_ratio = None
        for i, (ftr, ftr_count) in enumerate(self._index_ftr):
            if ftr == "motif3":
                # calculate { motif_index: motif ratio }
                motif3_ratio = self._count_subgraph_motif_by_size(self._ftr_mx, ftr) if not motif3_ratio else motif3_ratio
                final_vec[0, i] = motif3_ratio[ftr_count]
                self._headers.append("motif3_" + str(self._motif_index_to_edge_num["motif3"][ftr_count]) + "_edges")
            elif ftr == "motif4":
                # calculate { motif_index: motif ratio }
                motif4_ratio = self._count_subgraph_motif_by_size(self._ftr_mx, ftr) if not motif4_ratio else motif4_ratio
                final_vec[0, i] = motif4_ratio[ftr_count]
                self._headers.append("motif4_" + str(self._motif_index_to_edge_num["motif4"][ftr_count]) + "_edges")
            else:
                # calculate average of column
                final_vec[0, i] = np.sum(self._ftr_mx[:, i]) / self._ftr_mx.shape[0]
                self._headers.append(ftr + "_" + str(ftr_count))
        return final_vec

    def _build_matrix(self):
        sum_dictionaries_motifs3 = []
        sum_dictionaries_motifs4 = []
        # 3: [ ... row(node): { num_edges_in_motif3: count (for_this_node_only) } ... ]
        for i in range(self._ftr_mx.shape[0]):
            sum_dictionaries_motifs3.append({})
            sum_dictionaries_motifs4.append({})
            for j, (ftr, ftr_count) in enumerate(self._index_ftr):
                if ftr == "motif3":
                    key = self._motif_index_to_edge_num[ftr][ftr_count]
                    sum_dictionaries_motifs3[i][key] = sum_dictionaries_motifs3[i].get(key, 1e-3) + self._ftr_mx[i, j]
                elif ftr == "motif4":
                    key = self._motif_index_to_edge_num[ftr][ftr_count]
                    sum_dictionaries_motifs4[i][key] = sum_dictionaries_motifs4[i].get(key, 1e-3) + self._ftr_mx[i, j]

        return_mx = self._ftr_mx.copy()
        for i in range(self._ftr_mx.shape[0]):
            for j, (ftr, ftr_count) in enumerate(self._index_ftr):
                if ftr == "motif3":
                    # calculate { motif_index: motif ratio }
                    return_mx[i, j] /= sum_dictionaries_motifs3[i][self._motif_index_to_edge_num[ftr][ftr_count]]
                    if i == 0:
                        self._headers.append("motif3_" + str(self._motif_index_to_edge_num["motif3"][ftr_count]) + "_edges")
                elif ftr == "motif4":
                    # calculate { motif_index: motif ratio }
                    return_mx[i, j] /= sum_dictionaries_motifs4[i][self._motif_index_to_edge_num[ftr][ftr_count]]
                    if i == 0:
                        self._headers.append("motif4_" + str(self._motif_index_to_edge_num["motif4"][ftr_count]) + "_edges")
                else:
                    if i == 0:
                        self._headers.append(ftr + "_" + str(ftr_count))
        return return_mx

    def motif_ratio_vector(self):
        self._motif_ratio_vec = self._motif_ratio_vec if self._motif_ratio_vec else self._build_vector()
        return self._motif_ratio_vec[0]

    def motif_ratio_matrix(self):
        self._motif_ratio_matrix = self._motif_ratio_matrix if self._motif_ratio_matrix else self._build_matrix()
        return self._motif_ratio_matrix

    def get_headers(self):
        return self._headers

    # return { motif_index: sum motif in index/ total motifs with same edge count }
    def _count_subgraph_motif_by_size(self, ftr_mat, motif_type):
        sum_dict = {ftr_count: np.sum(ftr_mat[:, i]) for i, (ftr, ftr_count) in enumerate(self._index_ftr)
                    if ftr == motif_type}       # dictionary { motif_index: sum column }
        sum_by_edge = {}                        # dictionary { num_edges_in_motif: sum of  }
        for motif_count, sum_motif in sum_dict.items():
            key = self._motif_index_to_edge_num[motif_type][motif_count]
            sum_by_edge[key] = sum_by_edge.get(key, 0) + sum_motif
        # rewrite dictionary { motif_index: sum column/ total motifs with same edge count }
        for motif_count in sum_dict:
            key = self._motif_index_to_edge_num[motif_type][motif_count]
            sum_dict[motif_count] = sum_dict[motif_count] / sum_by_edge[key] if sum_by_edge[key] else 0
        return sum_dict

    # return [ ... (motif_type, counter) ... ]
    def _get_motif_type(self, motif_type, num_motifs):
        header = []
        for i in range(num_motifs):
            header.append((motif_type, i))
        return header

    @staticmethod
    def is_motif(ftr):
        return ftr == 'motif4' or ftr == "motif3"


if __name__ == "__main__":
    import networkx as nx
    from feature_meta import NODE_FEATURES

    gnx = nx.Graph()
    gnx.add_edges_from([
        (1, 2),
        (1, 3),
        (2, 3),
        (2, 7),
        (7, 8),
        (3, 6),
        (4, 6),
        (6, 8),
        (5, 6),
    ])
    gnx_ftr = GraphFeatures(gnx, NODE_FEATURES, "", is_max_connected=True)
    gnx_ftr.build()
    m = MotifRatio(gnx_ftr, False)
    e = 0



from dataclasses import asdict

from torch.utils.data import DataLoader

from QGCN.params import GraphsDataParams, ExternalParams, ModelParams, ActivatorParams
from QGCN.dataset.dataset_graphs_model import GraphsDataset
from QGCN.dataset.dataset_external_data import ExternalData
from QGCN.QGCN_model.qgcn_activator import QGCNActivator
from QGCN.QGCN_model.QGCN import QGCN


class QGCNDataSet:
    """
    This class gets the parameters of the data (as external_params and graph_data_params objects),
    and build from them datasets.
    """
    def __init__(self, dataset_name: str,
                 graph_data_params: GraphsDataParams = None,
                 external_params: ExternalParams = None,
                 params_file: str = None):

        if params_file:
            self.params = params_file
        elif graph_data_params and external_params:
            self.params = {
                "dataset_name": dataset_name,
                "external": asdict(external_params),
                "graphs_data": asdict(graph_data_params)
            }
        else:
            raise Exception('QGCNDataset should gets a params file or 2 params objects.')

        ext_train = ExternalData(self.params)
        self.dataset = GraphsDataset(self.params, external_data=ext_train)

    def get_dataset(self):
        return self.dataset


class QGCNModel:
    """
    This class create a dataset and a pytorch model of QGCN.
    You should init the model with the params dataclasses, and then call the train method.
    """
    def __init__(self, dataset_name: str,
                 graph_data_params: GraphsDataParams = None,
                 external_params: ExternalParams = None,
                 model_params: ModelParams = None,
                 activator_params: ActivatorParams = None,
                 device="cpu", params_file: str = None):

        if params_file:
            self.params = params_file
        elif graph_data_params and external_params and model_params and activator_params:
            self.params = {
                "dataset_name": dataset_name,
                "external": asdict(external_params),
                "graphs_data": asdict(graph_data_params),
                "model": asdict(model_params),
                "activator": asdict(activator_params)
            }
        else:
            raise Exception('QGCNModel should gets a params-file-path or 4 params objects.')

        self.ext_train = ExternalData(self.params)
        self.ds = GraphsDataset(self.params, external_data=self.ext_train)
        qgcn = QGCN(self.params, self.ds.len_features, self.ext_train.len_embed())
        self.activator = QGCNActivator(qgcn, self.params, self.ds, device=device)

    def get_dataset(self):
        return self.ds

    def train(self, should_print=True):
        self.activator.train(should_print=should_print)

    def predictl(self, dataloader: DataLoader):
        outputs = []
        model = self.activator.model
        model.eval()
        for _, (A, x0, embed) in enumerate(dataloader):
            outputs.append(model(A, x0, embed))

        return outputs

    def predict(self, A, x0, embed):
        model = self.activator.model
        model.eval()
        return model(A, x0, embed)

    def get_model(self):
        return self.activator.model



from _datetime import datetime
from graphMeasures.loggers import PrintLogger
from enum import Enum
import pandas as pd
from datetime import timedelta
from ..multi_graph.multi_graph import MultiGraph


class TemporalGraph:
    def __init__(self, database_name, csv_source, time_format, time_col, src_col, dst_col, weight_col=None,
                 weeks=0, days=0, hours=0, minutes=0, seconds=0, logger=None, time_format_out=None, directed=False):
        self._csv_source = csv_source
        self._database_name = database_name
        self._directed = directed
        self._time_format = time_format
        self._time_col = time_col
        self._src_col = src_col
        self._dst_col = dst_col
        self._weight_col = weight_col
        self._format_out = time_format_out if time_format_out else time_format
        self._timedelta = timedelta(weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds)
        self._time_col = time_col
        self._src_col = src_col
        self._logger = logger if logger else PrintLogger()
        self._mg_dictionary = self._break_by_time()

    def to_multi_graph(self):
        mg = MultiGraph(self._database_name, graphs_source=self._mg_dictionary, directed=self._directed)
        mg.sort_by(lambda x: datetime.strptime(x, self._format_out))
        return mg

    def _break_by_time(self):
        dict_time_edge_list = {}

        # open basic data csv (with all edges of all times)
        data_df = pd.read_csv(self._csv_source)
        self._format_data(data_df)
        # make TIME column the index column and sort data by it
        # data_df = data_df.set_index(self._time_col).sort_index()
        data_df = data_df.sort_values(self._time_col)

        # first day is the first row
        curr_time = data_df[self._time_col][data_df.first_valid_index()]
        # file_time = open(os.path.join(OUT_TMP_PATH, curr_day.date.strftime(format_out)), "wt")
        # next day is the  floor(<current_day>) + <one_time_interval>
        prev_time = curr_time
        next_time = prev_time + self._timedelta
        for index, row in data_df.iterrows():
            curr_time = data_df[self._time_col][index]
            # if curr day is in next time interval
            if curr_time >= next_time:
                while curr_time >= next_time:
                    prev_time = next_time
                    next_time = prev_time + self._timedelta

            # write edge to file
            src = str(row[self._src_col])
            dst = str(row[self._dst_col])
            weight = float(row[self._weight_col]) if self._weight_col else 1
            key = prev_time.strftime(self._format_out)
            if key not in dict_time_edge_list:
                dict_time_edge_list[key] = []
            dict_time_edge_list[key].append((src, dst, weight))
        return dict_time_edge_list

    def _format_data(self, graph_df):
        if self._time_format == "numerate":
            return
        if self._time_format == "MIL":
            graph_df[self._time_col] = graph_df[self._time_col] / 1000  # mili to seconds
        if self._time_format == "MIL" or format == "SEC":
            graph_df[self._time_col] = \
                graph_df[self._time_col].apply(lambda x: datetime.fromtimestamp(x))  # to datetime format
        else:
            # to datetime
            graph_df[self._time_col] = graph_df[self._time_col].apply(lambda x: datetime.strptime(str(x), self._time_format))


if __name__ == "__main__":
    pd.set_option('display.max_columns', 2)
    pd.set_option('display.max_rows', 2)
    pd.set_option('display.width', 1000)

    tg = TemporalGraph("FireWall", "Firewall-04062012.csv", '%d/%b/%Y %H%M:%S', "Date/time", "Source IP",
                       "Destination IP", minutes=2)
    e = 0



import numpy as np
from .motif_ratio import MotifRatio
from graphMeasures.features_meta.features_meta import FeaturesMeta
from graphMeasures.features_infra.graph_features import GraphFeatures


# Old zscore code.. should use scipy.stats.zscore
def z_scoring(matrix):
    new_matrix = np.asmatrix(matrix)
    minimum = np.asarray(new_matrix.min(0))  # column wise
    for i in range(minimum.shape[1]):
        if minimum[0, i] > 0:
            new_matrix[:, i] = np.log10(new_matrix[:, i])
        elif minimum[0, i] == 0:
            new_matrix[:, i] = np.log10(new_matrix[:, i] + 0.1)
        if new_matrix[:, i].std() > 0:
            new_matrix[:, i] = (new_matrix[:, i] - new_matrix[:, i].mean()) / new_matrix[:, i].std()
    return new_matrix


def log_norm(matrix):
    matrix[np.isnan(matrix)] = 1e-10
    matrix = np.abs(matrix)
    matrix[matrix < 1e-10] = 1e-10
    return np.log(matrix)


class FeaturesProcessor:
    def __init__(self, gnx_features: GraphFeatures):
        self._gnx_ftr = gnx_features
        if not self._gnx_ftr.is_build:
            self._gnx_ftr.build()

    # function input:
    #   - to add: list/np_array/np_matrix
    #   - norm_func: function that takes a vector and outputs processed outcome
    # the function calculates a vector with ratio-motifs and average features add <to_add> and activates the
    #     <norm_func> on the vector
    # function output: the output of the above
    def activate_motif_ratio_vec(self, to_add=None, norm_func=None):
        as_vec = MotifRatio(self._gnx_ftr, self._gnx_ftr.graph.is_directed()).motif_ratio_vector()
        if to_add:
            as_vec = np.hstack((as_vec, np.matrix(to_add)))
        if norm_func:
            as_vec = norm_func(as_vec)
        return np.array(as_vec)

    @staticmethod
    def _convert_dict_to_list(dictionary, entries_order):
        if type(dictionary) == dict:
            dict_as_list = []
            for vertex in entries_order:
                dict_as_list.append(dictionary[vertex]) if type(dictionary[vertex]) == list else \
                    dict_as_list.append([dictionary[vertex]])
            return dict_as_list
        return dictionary

    # function input:
    #   - to add: list/dictionary/np_matrix
    #   - norm_func: function that takes a matrix and outputs processed outcome
    # the function adds <to_add> to the ftr_matrix and activates the <norm_func> on the matrix
    # function output: the output of the above
    def as_matrix(self, entries_order: list = None, to_add=None, norm_func=None):
        entries_order = entries_order if entries_order else sorted(self._gnx_ftr.graph)
        as_matrix = self._gnx_ftr.to_matrix(entries_order=entries_order, dtype=np.float32, mtype=np.matrix,
                                            should_zscore=False)
        if to_add:
            to_add = self._convert_dict_to_list(to_add, entries_order)
            as_matrix = np.hstack((as_matrix, np.matrix(to_add)))
        if norm_func:
            as_matrix = norm_func(as_matrix)
        return as_matrix


if __name__ == "__main__":
    import networkx as nx

    gnx = nx.Graph()
    gnx.add_edges_from([
        (1, 2),
        (1, 3),
        (2, 3),
        (2, 7),
        (7, 8),
        (3, 6),
        (4, 6),
        (6, 8),
        (5, 6),
    ])
    gnx_ftr = GraphFeatures(gnx, FeaturesMeta().NODE_LEVEL, "", is_max_connected=True)
    fp = FeaturesProcessor(gnx_ftr)
    fp.activate_motif_ratio_vec()
    e = 0



"""
    all the graphs are stored under one graph
    - each edge has a set of attributes - {graph_1: weight_1, ...... graph_k: weight_k}
    - if the database files name are with dates, use date_format='format' for example date_format='%d-%b-%Y'
"""
import networkx as nx
import os
from graphMeasures.loggers import PrintLogger, BaseLogger


class MultiGraph:
    """
    database_name:  name of the data set
    source=

        None: in that case, it is expected that a dir named DATABASE_NAME will be in /GRAPHS_INPUT/multi_graph
              the directory should contain files, each file is a single graph, in the following format:
              each line is a single edge    source,dest,weight           # weight is optional
        PATH: same as above but the directory will be the given path
        DICTIONARY { name: gnx }
        DICTIONARY { name: edge_list=[...(source,dest,weight)...] }      # weight is optional
    """
    def __init__(self, database_name, graphs_source=None, directed=False, logger: BaseLogger=None):
        self._base_dir = __file__.replace("/", os.sep)
        self._base_dir = os.path.join(self._base_dir.rsplit(os.sep, 1)[0], "..")
        self._databese_name = database_name
        self._directed = directed
        self._gnx_multi = nx.DiGraph() if directed else nx.Graph()
        self._logger = logger if logger else PrintLogger("multi-graph")
        self._source = graphs_source

        self._list_id = []          # list of graph ID's - gives a mapping index to ID
        self._dict_id = {}          # dictionary of graph ID's - gives a mapping ID to index
        self._node_count = {}
        self._node_lists = {}
        self._edge_count = {}
        self._graph_valid = {}

        self._build(self._source)
        self._logger.debug("finish initialization")

    def suspend_logger(self):
        self._logger = None

    def wake_logger(self, logger: BaseLogger=None):
        self._logger = logger if logger else PrintLogger("multi-graph")

    def _build(self, source):
        if not source:
            source = os.path.join(self._base_dir, "GRAPHS_INPUT", "multi_graph", self._databese_name)
        if type(source) == str:
            self._build_from_files(source)
        elif type(source) == dict and len(source) > 0:
            item = next(iter(source.values()))
            if type(item) == nx.classes.graph.Graph or type(item) == nx.classes.digraph.DiGraph:
                self._build_from_gnx_dict(source)
            if type(item) == list and len(item) > 0 and type(item[0]) == tuple:
                self._build_from_edge_list_dict(source)
        else:
            self._logger.error("graph_source is not in the correct format")
            exit(1)

    @staticmethod
    def _strip_txt(file_name):
        return file_name.split(".")[0]

    @staticmethod
    def _split_row(row):
        row = row.split()
        edge = [1, 1, 1]
        for i, val in enumerate(row):
            edge[i] = val
        return edge

    def _add_edge_to_gnx(self, graph_name, node_u, node_v, weight):
        self._logger.debug("adding edge:\t(" + str(node_u) + "," + str(node_v) + ")\tweight=" + str(weight))

        # count edges
        curr_edge_count = self._edge_count.get(graph_name, 0)
        # counter++ if:
        #   1. the edge exists in the multi_graph but it doesnt belong to the this specific graph
        #   2. the edge is not in the multi graph at all
        if (self._gnx_multi.has_edge(node_u, node_v) and graph_name not in self._gnx_multi.edges[node_u, node_v]) or \
                not self._gnx_multi.has_edge(node_u, node_v):
            self._edge_count[graph_name] = curr_edge_count + 1

        # count nodes
        curr_node_count = self._node_count.get(graph_name, 0)
        curr_node_list = self._node_lists.get(graph_name, [])
        if node_u not in curr_node_list:
            curr_node_count += 1
            curr_node_list += [node_u]
        if node_v not in curr_node_list:
            curr_node_count += 1
            curr_node_list += [node_v]
        self._node_count[graph_name] = curr_node_count
        self._node_lists[graph_name] = curr_node_list

        # add the edge to the graph if it doesn't exist
        self._gnx_multi.add_edge(node_u, node_v)
        # add {graph_name=weight_under_the_graph} to the attributes of the edge
        self._gnx_multi.edges[node_u, node_v][graph_name] = float(weight)

    def _index_graph(self, graph_name, valid=True):
        # save graph indexing
        if graph_name not in self._dict_id:
            self._dict_id[graph_name] = len(self._list_id)
            self._list_id.append(graph_name)
            self._graph_valid[graph_name] = valid

    def _build_from_files(self, source):
        # check source file exist
        if not source or not os.path.exists(source):
            self._logger.error("\n\n\t\t*** THE PATH TO THE DATABASE FILES IS NOT VALID !!! ***\n")
            exit(1)

        for graph_name in sorted(os.listdir(source)):
            graph_file = open(os.path.join(source, graph_name))
            graph_name = self._strip_txt(graph_name)
            # save graph indexing
            self._index_graph(graph_name)
            for row in graph_file:
                [node_u, node_v, weight] = self._split_row(row)
                self._add_edge_to_gnx(graph_name, node_u, node_v, weight)

            graph_file.close()

    def _build_from_edge_list_dict(self, source):
        for graph_name, edge_list in source.items():
            # save graph indexing
            self._index_graph(graph_name)
            for edge in edge_list:
                node_u = edge[0]
                node_v = edge[1]
                weight = 1 if len(edge) < 3 else edge[2]
                self._add_edge_to_gnx(graph_name, node_u, node_v, weight)

    def _build_from_gnx_dict(self, source):
        for graph_name, gnx in source.items():
            # save graph indexing
            self._index_graph(graph_name)
            for edge in gnx.edges(data=True):
                node_u = edge[0]
                node_v = edge[1]
                weight = edge[2]['weight'] if 'weight' in edge[2] else 1
                self._add_edge_to_gnx(graph_name, node_u, node_v, weight)

    def _subgraph_by_name(self, graph_name: str):
        if not self.is_graph(graph_name):
            self._logger.error("no graph named:\t" + graph_name)
            return
        subgraph_edges = nx.DiGraph() if self._directed else nx.Graph()
        for edge in list(self._gnx_multi.edges(data=True)):
            if graph_name in edge[2]:
                # edge is saved in the following method (from, to, {graph_name_i: weight_in_graph_i})
                subgraph_edges.add_edge(edge[0], edge[1], weight=float(edge[2][graph_name]))
        return subgraph_edges

    def _subgraph_by_index(self, index: int):
        if index < 0 or index > len(self._list_id):
            self._logger.error("index is out of scope - index < number_of_graphs/ index > 0")
        return self._subgraph_by_name(self._list_id[index])

    @staticmethod
    def _graph_from_list_in_dict(names_list, names_dict):
        for name in names_list:
            if name in names_dict:
                return True
        return False

    def _combined_graph_by_names(self, names_list=None, combine_all=False):
        # case combine_all - the returned graph will combination of all graphs regardless to the names list
        if combine_all:
            names_list = self._list_id
        else:
            for graph_name in names_list:
                if not self.is_graph(graph_name):
                    self._logger.error("no graph named:\t" + graph_name)
                    return
        subgraph_edges = nx.DiGraph() if self._directed else nx.Graph()
        for edge in list(self._gnx_multi.edges(data=True)):
            if self._graph_from_list_in_dict(names_list, edge[2]):
                # edge is saved in the following method (from, to, {graph_name_i: weight_in_graph_i})
                subgraph_edges.add_edge(edge[0], edge[1])
        return subgraph_edges

    def _combined_graph_by_indexes(self, index_list=None, combine_all=False):
        for i, index in enumerate(index_list):
            if index < 0 or index > len(self._list_id):
                self._logger.error("index is out of scope - index < number_of_graphs/ index > 0")
                return
            index_list[i] = self._list_id[index]
        return self._combined_graph_by_names(index_list, combine_all)

    def add_edges(self, source):
        if source:
            self._build(source)

    def remove_edges(self, edge_lists: dict):
        self._node_lists = {}
        self._node_count = {}
        self._graph_valid = {}
        self._list_id = []
        self._dict_id = {}

        # iterate over all edges
        for edge in list(self._gnx_multi.edges(data=True)):
            # if the edge needs to be removed for some specific graph remove the graph name from edge dictionary
            for graph_name in list(edge[2].keys()):
                self._index_graph(graph_name)
                if graph_name in edge_lists and \
                        ((edge[0], edge[1]) in edge_lists[graph_name] or
                         (not self._directed and (edge[1], edge[0]) in edge_lists[graph_name])):
                    del edge[2][graph_name]
                    self._edge_count[graph_name] -= 1
                else:
                    # count nodes
                    curr_node_count = self._node_count.get(graph_name, 0)
                    curr_node_list = self._node_lists.get(graph_name, [])
                    if edge[0] not in curr_node_list:
                        self._node_count[graph_name] = curr_node_count + 1
                        self._node_lists[graph_name] = curr_node_list + [edge[0]]
                    if edge[1] not in curr_node_list:
                        self._node_count[graph_name] += 1
                        self._node_lists[graph_name] = curr_node_list + [edge[1]]

            if len(edge[2]) == 0:
                self._gnx_multi.remove_edge(edge[0], edge[1])
        self._gnx_multi.remove_nodes_from(list(nx.isolates(self._gnx_multi)))

    def filter(self, func, func_input='gnx'):
        self.wake_logger()
        for i, gid in enumerate(self._list_id):
            if func_input == 'gnx':
                self._graph_valid[gid] = func(self._subgraph_by_name(gid))
            if func_input == 'graph_name':
                self._graph_valid[gid] = func(gid)
            if func_input == 'index':
                self._graph_valid[gid] = func(i)
            self._logger.info("filter graph " + gid + " :" + str(self._graph_valid[gid]))

    def clear_filters(self):
        for i in self._list_id:
            self._graph_valid[i] = True

    def sort_by(self, key_func):
        new_order = [i for i in sorted(self._list_id, key=key_func)]
        new_dict_id_to_index = {}
        for i, graph_id in enumerate(new_order):
            new_dict_id_to_index[graph_id] = i
        self._list_id = new_order
        self._dict_id = new_dict_id_to_index

    def get_gnx(self, graph_id):
        return self._subgraph_by_index(graph_id) if type(graph_id) == int else self._subgraph_by_name(graph_id)

    def combined_gnx(self, id_list: list=None):
        # return all graphs together
        if not list:
            return self._combined_graph_by_names(combine_all=True)
        # return by list
        if len(id_list) == 0:
            return nx.DiGraph() if self._directed else nx.Graph()
        return self._combined_graph_by_names(names_list=id_list) if type(id_list[0]) == str \
            else self._combined_graph_by_indexes(index_list=id_list)

    def is_graph(self, graph_id):
        if type(graph_id) == str:
            return True if graph_id in self._dict_id else False
        return True if graph_id < self.number_of_graphs() else False

    def index_to_name(self, index):
        if index < 0 or index > len(self._list_id):
            self._logger.error("index is out of scope - index < number_of_graphs/ index > 0")
            return None
        return self._list_id[index]

    def name_to_index(self, graph_name):
        if not self.is_graph(graph_name):
            self._logger.error("no graph named:\t" + graph_name)
            return None
        return self._dict_id[graph_name]

    def node_count(self, graph_id=None):
        # return count for specific graph
        if graph_id is not None:
            return self._node_count[self.index_to_name(graph_id)] if type(graph_id) is int else self._node_count[graph_id]
        # return list of edge count for all graphs
        return [self._node_count[i] for i in self._list_id]

    def edge_count(self, graph_id=None):
        # return count for specific graph
        if graph_id is not None:
            return self._edge_count[self.index_to_name(graph_id)] if type(graph_id) is int else self._edge_count[graph_id]
        # return list of edge count for all graphs
        return [self._edge_count[i] for i in self._list_id]

    def graphs(self, start_id=None, end_id=None):
        for gid in self._list_id[start_id: end_id]:
            if self._graph_valid[gid]:
                yield self._subgraph_by_name(gid)

    def graph_names(self, start_id=None, end_id=None):
        for gid in self._list_id[start_id: end_id]:
            if self._graph_valid[gid]:
                yield gid

    def number_of_graphs(self):
        return len(self._list_id)


def multi_graph_testing():
    mg = MultiGraph("EnronInc_by_day")
    mg.filter(lambda x: False if x.number_of_nodes() < 1500 else True)
    e = 0


if __name__ == "__main__":
    # pass
    multi_graph_testing()



import sys
import os
f = open("curr_pwd", "wt")
cwd = os.getcwd()
f.write(cwd)
f.close()

sys.path.insert(1, os.path.join(cwd, "../../.."))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures", "features_algorithms"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures", "graph_infra"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures", "features_infra"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures", "features_meta"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graph-measures", "features_algorithms", "vertices"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graphs-package", "features_processor"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graphs-package", "multi_graph"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graphs-package", "temporal_graphs"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graphs-package", "features_processor", "motif_variations"))
sys.path.insert(1, os.path.join(cwd, "../../..", "..", "graphs-package"))
import json
import ast
import pytorch_lightning as pl

from QGCN.QGCN_model.qgcn_activator import QGCNActivator
from QGCN.QGCN_model.QGCN import QGCN
from QGCN.dataset.dataset_external_data import ExternalData
from QGCN.dataset.dataset_graphs_model import GraphsDataset
from datetime import datetime


pl.seed_everything(0)


def create_params_dict(params_file, features, lr, dropout, l2, gcn_layers, standardization):
    dict_params = json.load(open(params_file, "rt"))
    dict_params["graphs_data"]["features"] = features
    dict_params["graphs_data"]["standardization"] = standardization
    dict_params["model"]["dropout"] = dropout
    dict_params["model"]["lr"] = lr
    dict_params["model"]["L2_regularization"] = l2
    dict_params["model"]["GCN_layers"] = ast.literal_eval(gcn_layers)
    return dict_params
    

def create_params_dict2(params_file, adj_norm, lr, dropout, l2, gcn_layers, standardization):
    dict_params = json.load(open(params_file, "rt"))
    dict_params["graphs_data"]["standardization"] = standardization
    dict_params["model"]["dropout"] = dropout
    dict_params["model"]["lr"] = lr
    dict_params["model"]["L2_regularization"] = l2
    dict_params["graphs_data"]["adjacency_norm"] = adj_norm
    dict_params["model"]["GCN_layers"] = ast.literal_eval(gcn_layers)
    return dict_params


def main1(params_file, device, name, fr, i):
    curr_time = datetime.now().strftime("%Y%m%d_%H%M")
    csv_file = "../grid_results/{}/{} grid search results_{} {}.csv".format(name, name, i, curr_time)
    csv_columns = ["features", "lr", "dropout", "l2", "layer1", "layer2", "standardization", "acc"]
    
    f = open(csv_file, "wt")
    header = csv_columns
    f.write(",".join(header) + "\n")
    f.close()
    
    #features = [["DEG", "CENTRALITY", "BFS"], ["DEG", "BFS"], ["BFS", "CENTRALITY"], ["DEG", "CENTRALITY"]]
    lrrr = [5e-3, 5e-4, 1e-4]
    dropout = [0.2, 0.5]
    l2 = [5e-3, 1e-4, 5e-4]
    gcn_layers = ["[{\"in_dim\": \"None\", \"out_dim\": 250},{\"in_dim\": 250, \"out_dim\": 100}]", "[{\"in_dim\": \"None\", \"out_dim\": 500},{\"in_dim                        \": 500, \"out_dim\": 250}]"]
    standardization = ["min_max", "zscore"]
    batch_size = 64
    optimizer = "ADAM_"
    activation = "tanh_"
    adjacency_norm = "NORM_REDUCED_SYMMETRIC"
    
    for lr in lrrr:
        for regr in l2:
            for d in dropout:
                for gl in gcn_layers:
                    for std in standardization:
                        dict_params = create_params_dict(params_file, fr, lr, d, regr, gl, std)
                        if name == "mutagen" or name=="grec":
                            ext_train = ExternalData(params_file)
                            ds = GraphsDataset(params_file, external_data=ext_train)
                            model = QGCN(params_file, ds.len_features, ext_train.len_embed())
                        else:
                            ds = GraphsDataset(params_file, external_data=None)
                            model = QGCN(params_file, ds.len_features, [10])
                        activator = QGCNActivator(model, params_file, ds, device=device, grid=True)
                        acc = activator.train()
                        f = open(csv_file, "a")
                        layer1 = gl.split(":")[2].split(")")[0].split("}")[0].split(" ")[1]
                        layer2 = gl.split(":")[4].split("}")[0].split(" ")[1]
                        my_f = "".join(fr[i] + ":"for i in range(len(fr)))
                        data = {"features": my_f, "lr": lr, "dropout": d, "l2": regr, "layer1": layer1, "layer2": layer2, "standardization": std, "acc":                                     acc}
                        f.write(",".join([str(x) for x in list(data.values())]) + "\n")  # parameters
                        f.close()
                        

def main2(params_file, device, name, lr, i):
    curr_time = datetime.now().strftime("%Y%m%d_%H%M")
    csv_file = "../grid_results/{}/{} grid search results_{} {}.csv".format(name, name, i, curr_time)
    csv_columns = ["adj_norm", "lr", "dropout", "l2", "layer1", "layer2", "standardization", "acc"]
    
    f = open(csv_file, "wt")
    header = csv_columns
    f.write(",".join(header) + "\n")
    f.close()
    
    #lrrr = [5e-3, 5e-4, 1e-4]
    dropout = [0.2, 0.5]
    l2 = [5e-3, 1e-4, 5e-4]
    gcn_layers = ["[{\"in_dim\": \"None\", \"out_dim\": 250},{\"in_dim\": 250, \"out_dim\": 250}]", "[{\"in_dim\": \"None\", \"out_dim\": 200},{\"in_dim                        \": 200, \"out_dim\": 100}]"]
    standardization = ["min_max", "zscore"]
    batch_size = 64
    optimizer = "ADAM_"
    activation = "tanh_"
    adjacency_norm = ["NORM_REDUCED_SYMMETRIC", "NORM_REDUCED"]
    
    for regr in l2:
        for d in dropout:
            for gl in gcn_layers:
                for std in standardization:
                    for adj_norm in adjacency_norm:
                        dict_params = create_params_dict2(params_file, adj_norm, lr, d, regr, gl, std)
                        if name == "mutagen" or name=="grec":
                            ext_train = ExternalData(params_file)
                            ds = GraphsDataset(params_file, external_data=ext_train)
                            model = QGCN(params_file, ds.len_features, ext_train.len_embed())
                        else:
                            ds = GraphsDataset(params_file, external_data=None)
                            model = QGCN(params_file, ds.len_features, [10])
                        activator = QGCNActivator(model, params_file, ds, device=device, grid=True)
                        acc = activator.train()
                        f = open(csv_file, "a")
                        layer1 = gl.split(":")[2].split(")")[0].split("}")[0].split(" ")[1]
                        layer2 = gl.split(":")[4].split("}")[0].split(" ")[1]
                        data = {"adj_norm": adj_norm, "lr": lr, "dropout": d, "l2": regr, "layer1": layer1, "layer2": layer2, "standardization": std,                                           "acc": acc}
                        f.write(",".join([str(x) for x in list(data.values())]) + "\n")  # parameters
                        f.close()
                        

def main(params_file, device, name, i):
    features = [["DEG", "CENTRALITY", "BFS"], ["DEG", "BFS"], ["BFS", "CENTRALITY"], ["DEG", "CENTRALITY"]]
    #lr = [5e-3, 5e-4, 1e-4]
    lr = [1e-3, 5e-3, 8e-5]
    if name == "mutagen":
        main1(params_file, device, name, features[i], i)
    else:
        main2(params_file, device, name, lr[i], i)
    



if __name__ == "__main__":
    device = "cuda:4"
    params_file = "../../../test/params/NCI1_params.json"
    main(params_file, device, "NCI1", 2)



from dataclasses import dataclass, field


def get_embeddings_dim():
    return [10, 10]


def get_GCN_layers():
    return [
        {"in_dim": "None", "out_dim": 100},
        {"in_dim": 100, "out_dim": 50},
        {"in_dim": 50, "out_dim": 25}
    ]


def get_default_features():
    return ["DEG", "CENTRALITY", "BFS"]


def get_embeddings():
    return ["chem", "symbol"]


def get_continuous():
    return ["charge", "x", "y"]


@dataclass
class ModelParams:
    label_type: str = "binary"
    num_classes: int = 2
    use_embeddings: str = "True"
    embeddings_dim: list[int] = field(default_factory=get_embeddings_dim)
    activation: str = "relu_"
    dropout: float = 0
    lr: float = 1e-3
    optimizer: str = "ADAM_"
    L2_regularization: float = 0
    f: str = "c_x0"
    GCN_layers: list[dict] = field(default_factory=get_GCN_layers)


@dataclass
class ActivatorParams:
    epochs: int = 3
    batch_size: int = 128
    loss_func: str = "binary_cross_entropy_with_logits_"
    train: float = 0.3467
    dev: float = 0.1153
    test: float = 0.538


@dataclass
class GraphsDataParams:
    file_path: str = "../../test/data/AIDS_all.csv"
    graph_col: str = "g_id"
    src_col: str = "src"
    dst_col: str = "dst"
    label_col: str = "label"
    directed: str = "False"
    features: list[str] = field(default_factory=get_default_features)
    adjacency_norm: str = "NORM_REDUCED"
    percentage: float = 1
    standardization: str = "zscore"


@dataclass
class ExternalParams:
    file_path: str = "../../test/data/AIDS_external_data_all.csv"
    graph_col: str = "g_id"
    node_col: str = "node"
    embeddings: list[str] = field(default_factory=get_embeddings)
    continuous: list[str] = field(default_factory=get_continuous)



"""
Here is the implementation of helpful class and functions that handle the given dataset.
"""
import os
import json
import sys

from scipy.stats import zscore
import scipy.sparse as sp
from torch import Tensor
from torch.nn import ConstantPad2d
from torch.utils.data import Dataset, DataLoader
from collections import Counter
import pandas as pd
import networkx as nx
import pickle
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from graphMeasures.features_algorithms.vertices.betweenness_centrality import BetweennessCentralityCalculator
from graphMeasures.features_algorithms.vertices.bfs_moments import BfsMomentsCalculator

# from graphs-package.features_processor.features_processor import FeaturesProcessor, log_nor
from QGCN.graphs_package.features_processor.features_processor import FeaturesProcessor, log_norm
from graphMeasures.features_infra.feature_calculators import FeatureMeta
from graphMeasures.features_infra.graph_features import GraphFeatures
from graphMeasures.loggers import PrintLogger
# from graphs_pach .multi_graph import MultiGraph
from QGCN.graphs_package.multi_graph.multi_graph import MultiGraph
from QGCN.dataset.dataset_external_data import ExternalData


# some important shortenings
PKL_DIR = "pkl"
NORM_REDUCED = "NORM_REDUCED"
NORM_REDUCED_SYMMETRIC = "NORM_REDUCED_SYMMETRIC"
IDENTITY = "IDENTITY"
RAW_FORM = "RAW_FORM"
DEG = "DEG"
IN_DEG = "IN_DEG"
OUT_DEG = "OUT_DEG"
CENTRALITY = ("betweenness_centrality", FeatureMeta(BetweennessCentralityCalculator, {"betweenness"}))
BFS = ("bfs_moments", FeatureMeta(BfsMomentsCalculator, {"bfs"}))


class GraphsDataset(Dataset):
    def __init__(self, params, external_data: ExternalData = None):
        self._params = params if type(params) is dict else json.load(open(params, "rt"))
        self._dataset_name = self._params["dataset_name"]
        self._params = self._params["graphs_data"]
        self._logger = PrintLogger("logger")
        if self._params["directed"] == "False":
            self._params["directed"] = False
        else:
            self._params["directed"] = True

        # the base dir is the dir where the main module is.
        self._base_dir = os.path.dirname(sys.modules['__main__'].__file__)

        self._external_data = external_data
        # init ftr_meta dictionary and other ftr attributes
        self._init_ftrs()
        self._src_file_path = os.path.join(self._params["file_path"])
        self._multi_graph, self._labels, self._label_to_idx, self._idx_to_label = self._build_multi_graph()
        self._data, self._idx_to_name = self._build_data()

    @property
    def all_labels(self):
        return self._idx_to_label

    @property
    def label_count(self):
        return Counter([v[3] for name, v in self._data.items()])

    def label(self, idx):
        return self._data[self._idx_to_name[idx]][3]

    @property
    def len_features(self):
        return self._data[self._idx_to_name[0]][1].shape[1]

    def _init_ftrs(self):
        self._deg, self._in_deg, self._out_deg, self._is_ftr, self._ftr_meta = False, False, False, False, {}
        self._is_external_data = False if self._external_data is None else True
        # params.FEATURES contains string and list of two elements (matching to key: value)
        # should Deg/In-Deg/Out-Deg be calculated
        for ftr in self._params["features"]:
            ftr = globals()[ftr]
            if ftr == DEG:
                self._deg = True
            elif ftr == IN_DEG:
                self._in_deg = True
            elif ftr == OUT_DEG:
                self._out_deg = True
            else:
                self._ftr_meta[ftr[0]] = ftr[1]

        # add directories for pickles
        if len(self._ftr_meta) > 0:
            self._ftr_path = os.path.join(self._base_dir, PKL_DIR, "ftr", self._dataset_name)
            if not os.path.exists(self._ftr_path):
                os.makedirs(self._ftr_path)
            self._is_ftr = True

    """
    build multi graph according to csv 
    each community is a single graph, no consideration to time
    """
    def _build_multi_graph(self):
        path_pkl = os.path.join(self._base_dir, PKL_DIR, self._dataset_name + "_split_" +
                                str(self._params["percentage"]) + "_mg.pkl")
        if os.path.exists(path_pkl):
            return pickle.load(open(path_pkl, "rb"))
        multi_graph_dict = {}
        labels = {}
        label_to_idx = {}
        # open basic data csv (with all edges of all times)
        data_df = pd.read_csv(self._src_file_path)
        stop = data_df.shape[0] * self._params["percentage"]

        for index, edge in data_df.iterrows():
            if index > stop:
                break
            # write edge to dictionary
            graph_id = str(edge[self._params["graph_col"]])
            src = str(edge[self._params["src_col"]])
            dst = str(edge[self._params["dst_col"]])
            multi_graph_dict[graph_id] = multi_graph_dict.get(graph_id, []) + [(src, dst)]
            label = edge[self._params["label_col"]]
            label_to_idx[label] = len(label_to_idx) if label not in label_to_idx else label_to_idx[label]
            labels[graph_id] = label_to_idx[label]

        mg = MultiGraph(self._dataset_name, graphs_source=multi_graph_dict,
                        directed=self._params["directed"], logger=self._logger)
        idx_to_label = [l for l in sorted(label_to_idx, key=lambda x: label_to_idx[x])]
        mg.suspend_logger()
        os.makedirs(os.path.join(self._base_dir, PKL_DIR), exist_ok=True)
        pickle.dump((mg, labels, label_to_idx, idx_to_label), open(path_pkl, "wb"))
        mg.wake_logger()
        return mg, labels, label_to_idx, idx_to_label

    """
    returns a vector x for gnx 
    basic version returns degree for each node
    """
    def _gnx_vec(self, gnx_id, gnx: nx.Graph, node_order):
        final_vec = []
        if self._deg:
            degrees = gnx.degree(gnx.nodes)
            final_vec.append(np.matrix([np.log(degrees[d] + 1e-3) for d in node_order]).T)
        if self._in_deg:
            degrees = gnx.in_degree(gnx.nodes)
            final_vec.append(np.matrix([np.log(degrees[d] + 1e-3) for d in node_order]).T)
        if self._out_deg:
            degrees = gnx.out_degree(gnx.nodes)
            final_vec.append(np.matrix([np.log(degrees[d] + 1e-3) for d in node_order]).T)
        if self._is_external_data and self._external_data.is_continuous:
            final_vec.append(np.matrix([self._external_data.continuous_feature(gnx_id, d) for d in node_order]))
        if self._is_ftr:
            name = str(gnx_id)
            gnx_dir_path = os.path.join(self._ftr_path, name)
            if not os.path.exists(gnx_dir_path):
                os.mkdir(gnx_dir_path)
            raw_ftr = GraphFeatures(gnx, self._ftr_meta, dir_path=gnx_dir_path, is_max_connected=False,
                                    logger=PrintLogger("logger"))
            raw_ftr.build(should_dump=True)  # build features
            final_vec.append(FeaturesProcessor(raw_ftr).as_matrix(norm_func=log_norm))

        return np.hstack(final_vec)

    def _degree_matrix(self, gnx, nodelist):
        degrees = gnx.degree(gnx.nodes)
        return np.diag([1/degrees[d] for d in nodelist])

    def _standardize_data(self, data):
        all_data_continuous_vec = []                # stack all vectors for all graphs
        key_to_idx_map = []                     # keep ordered list (g_id, num_nodes) according to stack order

        # stack
        for g_id, (A, gnx_vec, embed_vec, label) in data.items():
            all_data_continuous_vec.append(gnx_vec)
            key_to_idx_map.append((g_id, gnx_vec.shape[0]))  # g_id, number of nodes ... ordered
        all_data_continuous_vec = np.vstack(all_data_continuous_vec)

        # z-score data
        print(all_data_continuous_vec.shape)
        if self._params["standardization"] == "zscore":
            all_data_continuous_vec = np.array(all_data_continuous_vec)
            standardized_data = zscore(all_data_continuous_vec, axis=0)
        elif self._params["standardization"] == "scale":
            pass
        elif self._params["standardization"] == "min_max":
            scalar = MinMaxScaler()
            standardized_data = scalar.fit_transform(all_data_continuous_vec)

        # rebuild data to original form -> split stacked matrix according to <list: (g_id, num_nodes)>
        new_data_dict = {}
        start_idx = 0
        for g_id, num_nodes in key_to_idx_map:
            new_data_dict[g_id] = (data[g_id][0], standardized_data[start_idx: start_idx+num_nodes],
                                   data[g_id][2], data[g_id][3])
            start_idx += num_nodes

        return new_data_dict

    def _norm_adjacency(self, A, gnx, node_order):
        if self._params["adjacency_norm"] == NORM_REDUCED:
            # D^-0.5 A D^-0.5
            D = self._degree_matrix(gnx, nodelist=node_order)
            D_sqrt = np.matrix(np.sqrt(D))
            adjacency = D_sqrt * np.matrix(A) * D_sqrt

        elif self._params["adjacency_norm"] == NORM_REDUCED_SYMMETRIC:
            # D^-0.5 [A + A.T + I] D^-0.5
            D = self._degree_matrix(gnx, nodelist=node_order)
            D_sqrt = np.matrix(np.sqrt(D))
            adjacency = D_sqrt * np.matrix(A + A.T + np.identity(A.shape[0])) * D_sqrt
        elif self._params["adjacency_norm"] == IDENTITY:
            adjacency = np.identity(A.shape[0])
        elif self._params["adjacency_norm"] == RAW_FORM:
            adjacency = A
        else:
            print("Error in adjacency_norm: " + self._params["adjacency_norm"] + "is not a valid option")
            exit(1)
        return adjacency

    """
    builds a data dictionary
    { ... graph_name: ( A = Adjacency_matrix, x = graph_vec, label ) ... }  
    """
    def _build_data(self):
        ext_data_id = "None" if not self._is_external_data else "_embed_ftr_" + "_".join(self._external_data.embed_headers)\
                                                                + "_continuous_ftr_" + "_".join(self._external_data.continuous_headers) \
                                                                + "standardization_" + self._params["standardization"]
        pkl_path = os.path.join(self._base_dir, PKL_DIR, self._dataset_name + ext_data_id + "_data.pkl")
        if os.path.exists(pkl_path):
            return pickle.load(open(pkl_path, "rb"))
        data = {}
        idx_to_name = []

        for gnx_id, gnx in zip(self._multi_graph.graph_names(), self._multi_graph.graphs()):
            # if gnx.number_of_nodes() < 5:
            #     continue
            node_order = sorted(list(gnx.nodes()))
            idx_to_name.append(gnx_id)
            
            if self._params["adjacency_norm"] == NORM_REDUCED_SYMMETRIC:
                adjacency = nx.adjacency_matrix(gnx, nodelist=node_order).todense()
                #adjacency = self._norm_adjacency(nx.adjacency_matrix(gnx, nodelist=node_order).todense(), gnx, node_order)
                adj = sp.coo_matrix(adjacency)
                adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)
                adj = normalize(adj + sp.eye(adj.shape[0]))
                adjacency = adj.toarray()
            else:
                adjacency = self._norm_adjacency(nx.adjacency_matrix(gnx, nodelist=node_order).todense(), gnx, node_order)

            gnx_vec = self._gnx_vec(gnx_id, gnx, node_order)
            embed_vec = [self._external_data.embed_feature(gnx_id, d) for d in node_order] \
                if self._is_external_data and self._external_data.is_embed else None

            data[gnx_id] = (adjacency, gnx_vec, embed_vec, self._labels[gnx_id])

        data = self._standardize_data(data)
        pickle.dump((data, idx_to_name), open(pkl_path, "wb"))
        return data, idx_to_name

    def collate_fn(self, batch):
        lengths_sequences = []
        # calculate max word len + max char len
        for A, x, e, l in batch:
            lengths_sequences.append(A.shape[0])

        # in order to pad all batch to a single dimension max length is needed
        seq_max_len = np.max(lengths_sequences)

        # new batch variables
        adjacency_batch = []
        x_batch = []
        embeddings_batch = []
        labels_batch = []
        for A, x, e, l in batch:
            # pad word vectors
            adjacency_pad = ConstantPad2d((0, seq_max_len - A.shape[0], 0, seq_max_len - A.shape[0]), 0)
            adjacency_batch.append(adjacency_pad(A).tolist())
            vec_pad = ConstantPad2d((0, 0, 0, seq_max_len - A.shape[0]), 0)
            x_batch.append(vec_pad(x).tolist())
            embeddings_batch.append(vec_pad(e).tolist() if self._is_external_data and self._external_data.is_embed else e)
            labels_batch.append(l)

        return Tensor(adjacency_batch), Tensor(x_batch), Tensor(embeddings_batch).long(), Tensor(labels_batch).long()

    def __getitem__(self, index):
        gnx_id = self._idx_to_name[index]
        A, x, embed, label = self._data[gnx_id]
        embed = 0 if embed is None else Tensor(embed).long()
        return Tensor(A), Tensor(x), embed, label

    def __len__(self):
        return len(self._idx_to_name)


def normalize(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx


if __name__ == "__main__":
    ext_train = ExternalData("../params/default_binary_params.json")
    ds = GraphsDataset("../params/default_binary_params.json", external_data=ext_train)
    # ds = BilinearDataset(AidsDatasetTestParams())
    dl = DataLoader(
        dataset=ds,
        collate_fn=ds.collate_fn,
        batch_size=64,
    )
    p = []
    for i, (A, x, e, l) in enumerate(dl):
        print(i, A, x, e, l)
    e = 0



import copy
from datetime import timedelta, datetime
from graphMeasures.loggers import PrintLogger
import pandas as pd
from datetime import timedelta
from ..multi_graph.multi_graph import MultiGraph
from collections import Counter
import os


class TemporalMultiGraph:
    def __init__(self, database_name, csv_source, time_format, time_col, src_col, dst_col, subgraph_name_col,
                 weight_col=None, label_col=None, weeks=0, days=0, hours=0, minutes=0, seconds=0, logger=None,
                 time_format_out=None, directed=False):
        self._labels = {}
        self._order = {}
        self._times = []
        self._times_index = {}
        self._graphs_for_time = {}
        self._subgraph_name_col = subgraph_name_col
        self._csv_source = csv_source
        self._database_name = database_name
        self._directed = directed
        self._time_format = time_format
        self._time_col = time_col
        self._src_col = src_col
        self._dst_col = dst_col
        self._weight_col = weight_col
        self._label_col = label_col
        self._format_out = self._time_format_out(time_format_out)
        self._timedelta = timedelta(weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds)
        self._time_col = time_col
        self._src_col = src_col
        self._logger = logger if logger else PrintLogger()
        self._edge_list_dict = self._break_by_time()
        self._mg_dict = self._build_multi_graphs()
        self._number_of_times = len(self._edge_list_dict)

    @property
    def labels(self):
        return self._labels

    def _time_format_out(self, format):
        if format:
            return format
        if self._time_format == "MIL" or self._time_format == "SEC":
            return '%d_%b_%Y %H_%M_%S'
        return self._time_format

    def _break_by_time(self):
        index_order = 0
        time_index = 0
        dict_time_to_mg_dict = {}

        # open basic data csv (with all edges of all times)
        data_df = pd.read_csv(self._csv_source)
        self._format_data(data_df)
        # make TIME column the index column and sort data by it
        data_df = data_df.sort_values(self._time_col)

        # first day is the first row
        curr_time = data_df[self._time_col][data_df.first_valid_index()]
        # file_time = open(os.path.join(OUT_TMP_PATH, curr_day.date.strftime(format_out)), "wt")
        curr_time_mg_dict = {}
        # next day is the  floor(<current_day>) + <one_time_interval>
        prev_time = curr_time
        next_time = prev_time + self._timedelta
        for index, row in data_df.iterrows():
            curr_time = data_df[self._time_col][index]
            # if curr day is in next time interval
            if curr_time >= next_time:
                # save current time multi-graph_dictionary
                dict_time_to_mg_dict[prev_time.strftime(self._format_out)] = curr_time_mg_dict
                curr_time_mg_dict = {}
                self._times.append(prev_time.strftime(self._format_out))
                self._times_index[prev_time.strftime(self._format_out)] = time_index
                time_index += 1
                while curr_time >= next_time:
                    prev_time = next_time
                    next_time = prev_time + self._timedelta

            # write edge
            graph_name = str(row[self._subgraph_name_col])
            src = str(row[self._src_col])
            dst = str(row[self._dst_col])
            weight = float(row[self._weight_col]) if self._weight_col else 1
            if graph_name not in curr_time_mg_dict:
                curr_time_mg_dict[graph_name] = []
            curr_time_mg_dict[graph_name].append((src, dst, weight))
            # keep order of graphs
            if graph_name not in self._order:
                self._order[graph_name] = index_order
                index_order += 1
            if prev_time.strftime(self._format_out) not in self._graphs_for_time:
                self._graphs_for_time[prev_time.strftime(self._format_out)] = []
            if graph_name not in self._graphs_for_time[prev_time.strftime(self._format_out)]:
                self._graphs_for_time[prev_time.strftime(self._format_out)].append(graph_name)
            # add label
            if self._label_col:
                self._labels[graph_name] = self._labels.get(graph_name, []) + [row[self._label_col]]
        # fix labels to br the most common
        for name, labels_list in self._labels.items():
            self._labels[name] = Counter(labels_list).most_common()[0][0]
        return dict_time_to_mg_dict

    def _build_multi_graphs(self):
        time_mg_dict = {}
        for time, mg_dict in self._edge_list_dict.items():
            # create multi graph for each time and order the subgraphs according to the order dictionary (FIFO)
            mg = MultiGraph(time, graphs_source=mg_dict, directed=self._directed)
            mg.sort_by(self._func_order)
            time_mg_dict[time] = mg
        return time_mg_dict

    def _func_order(self, graph_name):
        return self._order[graph_name]

    def _format_data(self, graph_df):
        if self._time_format == "MIL":
            graph_df[self._time_col] = graph_df[self._time_col] / 1000  # mili to seconds
        if self._time_format == "MIL" or format == "SEC":
            graph_df[self._time_col] = \
                graph_df[self._time_col].apply(lambda x: datetime.fromtimestamp(x))  # to datetime format
        else:
            # to datetime
            graph_df[self._time_col] = graph_df[self._time_col].apply(lambda x: datetime.strptime(x, self._time_format))

    def multi_graphs_by_time(self, start=None, end=None):
        start = start if start else 0
        stop = end if end else self._number_of_times
        for i in range(start, stop):
            yield self._mg_dict[self._times[i]]

    def multi_graph_by_window(self, window_size=None, start_time=0):
        if start_time < 0 or start_time > self._number_of_times:
            self._logger.error("invalid start time = " + str(start_time) + ", total intervals = " + str(self._number_of_times))
            return
        # build base mg
        mg = MultiGraph(self._database_name + "window", graphs_source=self._edge_list_dict[self._times[0]],
                        directed=self._directed)
        for i in range(1, start_time):
            mg.add_edges(self._edge_list_dict[self._times[i]])

        window_size = window_size if window_size else self._number_of_times
        for i in range(start_time, self._number_of_times):
            mg.suspend_logger()
            temp = copy.deepcopy(mg)
            mg.wake_logger()
            yield temp

            to_remove = i - window_size
            if to_remove >= 0:
                mg.remove_edges(self._edge_list_dict[self._times[to_remove]])
            mg.add_edges(self._edge_list_dict[self._times[i]])


if __name__ == "__main__":
    _params = {
        'logger_name': "logger",
        # Data parameters
        'days_split': 1,
        'start_time': 10,
        'window_size': None,
        'database': 'Refael',
        'data_file_name': 'Refael_07_18.csv',  # should be in ../data/
        'date_format': "%Y-%m-%d",  # Refael
        'directed': True,
        'white_label': 1,
        # graph_measures + beta vectors parameters
        'max_connected': False,
        # ML- parameters
        'min_nodes': 10,
        # AL - parameters
        'batch_size': 2,
        'queries_per_time': 2,
        'eps': 0.01,
        'target_recall': 0.7,
        'reveal_target': 0.6,
        'ml_method': "XG_Boost"
    }
    _database_ = TemporalMultiGraph(_params['database'],
                                    _params['data_file_name'],
                                    time_format='MIL',
                                    time_col='StartTime',
                                    src_col='SourceID',
                                    dst_col='DestinationID',
                                    label_col='target',
                                    subgraph_name_col='Community',
                                    days=_params['days_split'],
                                    time_format_out=_params['date_format'],
                                    directed=_params['directed'])
    for multi_graph in _database_.multi_graph_by_window(_params['window_size'], _params['start_time']):
        None
    e = 0



import json
from torch.nn import Module, Linear, Dropout, Embedding, ModuleList, LogSoftmax
import torch
from torch.optim import Adam, SGD
from torch.utils.data import DataLoader
import math

ADAM_ = Adam
SGD_ = SGD
relu_ = torch.relu
sigmoid_ = torch.sigmoid
tanh_ = torch.tanh
srss_ = lambda x: 1

"""
given A, x0 : A=Adjacency_matrix, x0=nodes_vec
First_model => x1(n x k) = sigma( A(n x n) * x0(n x d) * W1(d x k) )
Bilinear_model => x2(1 x 1) = sigma( W2(1 x k) * trans(x1)(k x n) * A(n x n) * x0(n x d) * W1(d x 1) )
"""


class GCN(Module):
    def __init__(self, in_dim, out_dim, activation, dropout, init_layers="reset"):
        super(GCN, self).__init__()
        # useful info in forward function
        self._linear = Linear(in_dim, out_dim)
        if init_layers == "reset":
            self.reset_parameters(out_dim)
        else:
            self.init()
        self.reset_parameters(out_dim)
        self._activation = activation
        self._dropout = Dropout(p=dropout)
        self._gpu = True

    def reset_parameters(self, out_dim):
        stdv = 1. / math.sqrt(out_dim)
        self._linear.weight.data.uniform_(-stdv, stdv)

    def init(self):
        final_weight_tensor = torch.nn.init.xavier_normal_(self._linear.weight.data)
        self._linear.weight.data = final_weight_tensor

    def forward(self, A, x0):
        # Dropout layer
        x0 = self._dropout(x0)
        Ax = torch.matmul(A, x0)

        x = self._linear(Ax)
        x1 = self._activation(x) - 2 / (x ** 2 + 1)

        return x1


class QGCNLastLayer(Module):
    def __init__(self, left_in_dim, right_in_dim, out_dim, init_layers="reset"):
        super(QGCNLastLayer, self).__init__()
        # useful info in forward function
        self._left_linear = Linear(left_in_dim, 1)
        self._right_linear = Linear(right_in_dim, out_dim)
        if init_layers == "reset":
            self.reset_parameters(out_dim)
        else:
            self.init()
        self._gpu = False

    def reset_parameters(self, out_dim):
        stdv = 1. / math.sqrt(1)
        self._left_linear.weight.data.uniform_(-stdv, stdv)
        stdv_r = 1. / math.sqrt(out_dim)
        self._right_linear.weight.data.uniform_(-stdv_r, stdv_r)

    def init(self):
        final_weight_tensor_l = torch.nn.init.xavier_uniform_(self._left_linear.weight.data)
        self._left_linear.weight.data = final_weight_tensor_l
        final_weight_tensor_r = torch.nn.init.uniform_(self._right_linear.weight.data)
        self._right_linear.weight.data = final_weight_tensor_r

    def forward(self, A, x0, x1):
        # sigmoid( W2(1 x k) * trans(x1)(k x n) * A(n x n) * x0(n x d) * W1(d x 1) )
        x1_A = torch.matmul(x1.permute(0, 2, 1), A)
        W2_x1_A = self._left_linear(x1_A.permute(0, 2, 1))
        W2_x1_A_x0 = torch.matmul(W2_x1_A.permute(0, 2, 1), x0)
        W2_x1_A_x0_W3 = self._right_linear(W2_x1_A_x0)
        return W2_x1_A_x0_W3.squeeze(dim=1)


class QGCN(Module):
    """
    first linear layer is executed numerous times
    """

    def __init__(self, params, in_dim, embed_vocab_dim, init_layers="reset"):
        super(QGCN, self).__init__()
        self._params = params["model"] if type(params) is dict else json.load(open(params, "rt"))["model"]

        # add embedding layers
        self._is_binary = True if self._params["label_type"] == "binary" else False
        self._is_embed = True if self._params["use_embeddings"] == "True" else False

        qgcn_layers_dim = [{"in": layer["in_dim"], "out": layer["out_dim"]} for layer in self._params["GCN_layers"]]
        qgcn_layers_dim[0]["in"] = in_dim

        if self._is_embed:
            # embeddings are added to ftr vector -> update dimensions of relevant weights
            qgcn_layers_dim[0]["in"] += sum(self._params["embeddings_dim"])

            # add embedding layers
            self._embed_layers = []
            for vocab_dim, embed_dim in zip(embed_vocab_dim, self._params["embeddings_dim"]):
                self._embed_layers.append(Embedding(vocab_dim, embed_dim))
            self._embed_layers = ModuleList(self._embed_layers)

        values_out = [qgcn_layers_dim[i]["out"] for i in range(len(qgcn_layers_dim))]
        values_out.append(qgcn_layers_dim[0]["in"])
        sum_layers_shape = sum(values_out)

        self._num_layers = len(self._params["GCN_layers"])
        self._linear_layers = []
        # create linear layers
        self._linear_layers = ModuleList([GCN(qgcn_layers_dim[i]["in"], qgcn_layers_dim[i]["out"],
                                              globals()[self._params["activation"]], self._params["dropout"],
                                              init_layers=init_layers) for i in range(self._num_layers)])
        if self._params["f"] == "x1_x0":
            qgcn_right_in = in_dim + sum(self._params["embeddings_dim"])
            qgcn_left_in = qgcn_layers_dim[-1]["out"]
        elif self._params["f"] == "x1_x1":
            qgcn_right_in = qgcn_layers_dim[-1]["out"]
            qgcn_left_in = qgcn_layers_dim[-1]["out"]
        elif self._params["f"] == "c_x1":
            qgcn_right_in = qgcn_layers_dim[-1]["out"]
            qgcn_left_in = sum_layers_shape
        else:
            qgcn_right_in = in_dim + sum(self._params["embeddings_dim"])
            qgcn_left_in = qgcn_layers_dim[-1]["out"]

        self._qgcn_last_layer = QGCNLastLayer(left_in_dim=qgcn_left_in, right_in_dim=qgcn_right_in,
                                              out_dim=1 if self._is_binary else self._params["num_classes"],
                                              init_layers=init_layers)
        self._softmax = LogSoftmax(dim=1)
        self.optimizer = self.set_optimizer(self._params["lr"], globals()[self._params["optimizer"]],
                                            self._params["L2_regularization"])

    def set_optimizer(self, lr, opt, weight_decay):
        return opt(self.parameters(), lr=lr, weight_decay=weight_decay)

    def _fix_shape(self, input):
        return input if len(input.shape) == 3 else input.unsqueeze(dim=0)

    def forward(self, A, x0, embed, test=False):
        if self._is_embed:
            list_embed = []
            for i, embedding in enumerate(self._embed_layers):
                list_embed.append(embedding(embed[:, :, i]))
            x0 = torch.cat([x0] + list_embed, dim=2)

        x1 = x0
        H_layers = [x1]
        for i in range(self._num_layers):
            x1 = self._linear_layers[i](A, x1)
            H_layers.append(x1)

        c = torch.cat(H_layers, dim=2)

        if self._params["f"] == "x1_x0":
            x2 = self._qgcn_last_layer(A, x0, x1)
        elif self._params["f"] == "x1_x1":
            x2 = self._qgcn_last_layer(A, x1, x1)
        elif self._params["f"] == "c_x1":
            x2 = self._qgcn_last_layer(A, x1, c)
        else:  # like "x1_x0"
            x2 = self._qgcn_last_layer(A, x0, x1)

        # x2 = self._qgcn_last_layer(A, x0, x1)
        if test:
            if self._is_binary:
                x2 = torch.sigmoid(x2)
            else:
                x2 = self._softmax(x2)
        return x2



import json
from sys import stdout
from time import sleep
from bokeh.io import output_file

import torch
from sklearn.metrics import roc_auc_score
from bokeh.plotting import figure, show
from torch.nn.functional import binary_cross_entropy_with_logits, cross_entropy
from torch.utils.data import DataLoader, random_split
from collections import Counter
import numpy as np
from QGCN.QGCN_model.QGCN import QGCN
from QGCN.dataset.dataset_graphs_model import GraphsDataset

TRAIN_JOB = "TRAIN"
DEV_JOB = "DEV"
TEST_JOB = "TEST"
VALIDATE_JOB = "VALIDATE"
LOSS_PLOT = "loss"
AUC_PLOT = "AUC"
ACCURACY_PLOT = "accuracy"

binary_cross_entropy_with_logits_ = binary_cross_entropy_with_logits
cross_entropy_ = cross_entropy

sigmoid = lambda x: 1 / (1 + np.exp(-x))


def clip_grads(model, clip_c):
    """
    Apply gradient clipping as the writers did in the paper's implementation:
    If a gradient is larger than clip_c ** 2, divide it by the 2-norm of all gradients vector times clip_c
    """
    total_norm = 0
    for p in filter(lambda x: x.requires_grad, model.parameters()):
        total_norm += (p.grad.data ** 2).sum()
    total_norm = total_norm ** 0.5
    for p in filter(lambda x: x.requires_grad, model.parameters()):
        p.grad.data = torch.where(p.grad.data > clip_c ** 2, p.grad.data.mul(1 / (clip_c * total_norm)), p.grad.data)


class QGCNActivator:
    def __init__(self, model: QGCN, params, train_data: GraphsDataset, nni=False, device="cuda:0", grid=False):
        self._nni = nni
        self.grid = grid
        self._params = params if type(params) is dict else json.load(open(params, "rt"))
        self._dataset_name = self._params["dataset_name"]
        self._is_binary = True if self._params["model"]["label_type"] == "binary" else False
        self._params = self._params["activator"]

        self._gpu = torch.cuda.is_available()
        # self._gpu = True
        self._device = torch.device(device if self._gpu else "cpu")
        print(self._device)
        self._model = model.to(device=self._device)
        self._epochs = self._params["epochs"]
        self._batch_size = self._params["batch_size"]
        self._loss_func = globals()[self._params["loss_func"]]
        self._load_data(train_data, self._params["train"], self._params["dev"], self._params["test"], self._batch_size)
        self._init_loss_and_acc_vec()
        self._init_print_att()

    # init loss and accuracy vectors (as function of epochs)
    def _init_loss_and_acc_vec(self):
        self._loss_vec_train = []
        self._loss_vec_dev = []
        self._loss_vec_test = []

        self._bar = 0.5
        self._accuracy_vec_train = []
        self._accuracy_vec_dev = []
        self._accuracy_vec_test = []

        self._auc_vec_train = []
        self._auc_vec_dev = []
        self._auc_vec_test = []

    # init variables that holds the last update for loss and accuracy
    def _init_print_att(self):
        self._print_train_accuracy = 0
        self._print_train_loss = 0
        self._print_train_auc = 0

        self._print_dev_accuracy = 0
        self._print_dev_loss = 0
        self._print_dev_auc = 0

        self._print_test_accuracy = 0
        self._print_test_loss = 0
        self._print_test_auc = 0

    # update loss after validating
    def _update_loss(self, loss, job=TRAIN_JOB):
        if job == TRAIN_JOB:
            self._loss_vec_train.append(loss)
            self._print_train_loss = loss
        elif job == DEV_JOB:
            self._loss_vec_dev.append(loss)
            self._print_dev_loss = loss
        elif job == TEST_JOB:
            self._loss_vec_test.append(loss)
            self._print_test_loss = loss

    # update accuracy after validating
    def _update_auc(self, pred, true, job=TRAIN_JOB):
        pred_ = [-1 if np.isnan(x) else x for x in pred]
        num_classes = len(Counter(true))
        if num_classes < 2:
            auc = 0.5
        # calculate acc
        else:
            auc = roc_auc_score(true, pred_)
        if job == TRAIN_JOB:
            self._print_train_auc = auc
            self._auc_vec_train.append(auc)
            return auc
        elif job == DEV_JOB:
            self._print_dev_auc = auc
            self._auc_vec_dev.append(auc)
            return auc
        elif job == TEST_JOB:
            self._print_test_auc = auc
            self._auc_vec_test.append(auc)
            return auc

    # update accuracy after validating
    def _update_accuracy_binary(self, pred, true, job=TRAIN_JOB):
        # calculate acc
        if job == TRAIN_JOB:
            max_acc = 0
            best_bar = self._bar
            for bar in [i * 0.01 for i in range(100)]:
                acc = sum([1 if (0 if i < bar else 1) == int(j) else 0 for i, j in zip(pred, true)]) / len(pred)
                if acc > max_acc:
                    best_bar = bar
                    max_acc = acc
            self._bar = best_bar

            self._print_train_accuracy = max_acc
            self._accuracy_vec_train.append(max_acc)
            return max_acc

        acc = sum([1 if (0 if i < self._bar else 1) == int(j) else 0 for i, j in zip(pred, true)]) / len(pred)

        if job == DEV_JOB:
            self._print_dev_accuracy = acc
            self._accuracy_vec_dev.append(acc)
            return acc
        elif job == TEST_JOB:
            self._print_test_accuracy = acc
            self._accuracy_vec_test.append(acc)
            return acc

    def _update_accuracy_multiclass(self, pred, true, job=TRAIN_JOB):
        pred = np.asarray(pred).argmax(axis=1).tolist()
        # calculate acc
        acc = sum([1 if int(i) == int(j) else 0 for i, j in zip(pred, true)]) / len(pred)
        if job == TRAIN_JOB:
            self._print_train_accuracy = acc
            self._accuracy_vec_train.append(acc)
            return acc
        elif job == DEV_JOB:
            self._print_dev_accuracy = acc
            self._accuracy_vec_dev.append(acc)
            return acc
        elif job == TEST_JOB:
            self._print_test_accuracy = acc
            self._accuracy_vec_test.append(acc)
            return acc

    # print progress of a single epoch as a percentage
    def _print_progress(self, batch_index, len_data, job=""):
        prog = int(100 * (batch_index + 1) / len_data)
        stdout.write("\r\r\r\r\r\r\r\r" + job + " %d" % prog + "%")
        print("", end="\n" if prog == 100 else "")
        stdout.flush()

    # print last loss and accuracy
    def _print_info(self, jobs=()):
        if self._is_binary:
            if TRAIN_JOB in jobs:
                print("Acc_Train: " + '{:{width}.{prec}f}'.format(self._print_train_accuracy, width=6, prec=4) +
                      " || AUC_Train: " + '{:{width}.{prec}f}'.format(self._print_train_auc, width=6, prec=4) +
                      " || Loss_Train: " + '{:{width}.{prec}f}'.format(self._print_train_loss, width=6, prec=4),
                      end=" || ")
            if DEV_JOB in jobs:
                print("Acc_Dev: " + '{:{width}.{prec}f}'.format(self._print_dev_accuracy, width=6, prec=4) +
                      " || AUC_Dev: " + '{:{width}.{prec}f}'.format(self._print_dev_auc, width=6, prec=4) +
                      " || Loss_Dev: " + '{:{width}.{prec}f}'.format(self._print_dev_loss, width=6, prec=4),
                      end=" || ")
            if TEST_JOB in jobs:
                print("Acc_Test: " + '{:{width}.{prec}f}'.format(self._print_test_accuracy, width=6, prec=4) +
                      " || AUC_Test: " + '{:{width}.{prec}f}'.format(self._print_test_auc, width=6, prec=4) +
                      " || Loss_Test: " + '{:{width}.{prec}f}'.format(self._print_test_loss, width=6, prec=4),
                      end=" || ")
            print("")
        else:
            if TRAIN_JOB in jobs:
                print("Acc_Train: " + '{:{width}.{prec}f}'.format(self._print_train_accuracy, width=6, prec=4) +
                      " || Loss_Train: " + '{:{width}.{prec}f}'.format(self._print_train_loss, width=6, prec=4),
                      end=" || ")
            if DEV_JOB in jobs:
                print("Acc_Dev: " + '{:{width}.{prec}f}'.format(self._print_dev_accuracy, width=6, prec=4) +
                      " || Loss_Dev: " + '{:{width}.{prec}f}'.format(self._print_dev_loss, width=6, prec=4),
                      end=" || ")
            if TEST_JOB in jobs:
                print("Acc_Test: " + '{:{width}.{prec}f}'.format(self._print_test_accuracy, width=6, prec=4) +
                      " || Loss_Test: " + '{:{width}.{prec}f}'.format(self._print_test_loss, width=6, prec=4),
                      end=" || ")
            print("")

    # plot loss / accuracy graph
    def plot_line(self, job=LOSS_PLOT, show_plot=True):
        p = figure(plot_width=600, plot_height=250, title=self._dataset_name + " - Dataset - " + job,
                   x_axis_label="epochs", y_axis_label=job)
        color1, color2, color3 = ("yellow", "orange", "red") if job == LOSS_PLOT else ("black", "green", "blue")
        if job == LOSS_PLOT:
            y_axis_train = self._loss_vec_train
            y_axis_dev = self._loss_vec_dev
            y_axis_test = self._loss_vec_test
        elif job == AUC_PLOT:
            y_axis_train = self._auc_vec_train
            y_axis_dev = self._auc_vec_dev
            y_axis_test = self._auc_vec_test
        elif job == ACCURACY_PLOT:
            y_axis_train = self._accuracy_vec_train
            y_axis_dev = self._accuracy_vec_dev
            y_axis_test = self._accuracy_vec_test

        x_axis = list(range(len(y_axis_dev)))
        p.line(x_axis, y_axis_train, line_color=color1, legend="train")
        p.line(x_axis, y_axis_dev, line_color=color2, legend="dev")
        p.line(x_axis, y_axis_test, line_color=color3, legend="test")
        if show_plot:
            output_file("foo.html")
            show(p)
        return p

    def _plot_acc_dev(self):
        self.plot_line(ACCURACY_PLOT)
        sleep(1)
        self.plot_line(ACCURACY_PLOT)
        sleep(1)
        self.plot_line(ACCURACY_PLOT)

    @property
    def model(self):
        return self._model

    @property
    def loss_train_vec(self):
        return self._loss_vec_train

    @property
    def accuracy_train_vec(self):
        return self._accuracy_vec_train

    @property
    def auc_train_vec(self):
        return self._auc_vec_train

    @property
    def loss_dev_vec(self):
        return self._loss_vec_dev

    @property
    def accuracy_dev_vec(self):
        return self._accuracy_vec_dev

    @property
    def auc_dev_vec(self):
        return self._auc_vec_dev

    @property
    def loss_test_vec(self):
        return self._loss_vec_test

    @property
    def accuracy_test_vec(self):
        return self._accuracy_vec_test

    @property
    def auc_test_vec(self):
        return self._auc_vec_test

    # load dataset
    def _load_data(self, dataset, train_size, dev_size, test_size, batch_size):
        # calculate lengths off train and dev according to split ~ (0,1)
        len_train = int(len(dataset) * train_size)
        len_dev = int(len(dataset) * dev_size)
        len_test = len(dataset) - len_train - len_dev

        # split dataset
        train, dev, test = random_split(dataset, (len_train, len_dev, len_test))

        # set train loader
        self._train_loader = DataLoader(
            train,
            batch_size=batch_size,
            collate_fn=train.dataset.collate_fn,
            shuffle=True
        )
        self._loss_weights = torch.Tensor([1 / count for label, count in
                                           sorted(list(dataset.label_count.items()), key=lambda x: x[0])])
        # set validation loader
        self._dev_loader = DataLoader(
            dev,
            batch_size=batch_size,
            collate_fn=dev.dataset.collate_fn,
        )

        # set test loader
        self._test_loader = DataLoader(
            test,
            batch_size=batch_size,
            collate_fn=test.dataset.collate_fn,
        )

    # train a model, input is the enum of the model type
    def train(self, show_plot=False, early_stop=False, wait=10, mul=0.5, stop_sign=8, norm_type=1, wd=0.001, should_print=True):
        lr = self._model._params["lr"]
        print(lr)
        wait_counter = 0
        bad_counter = 0
        self._init_loss_and_acc_vec()
        # calc number of iteration in current epoch
        len_data = len(self._train_loader)
        for epoch_num in range(self._epochs):
            if not self._nni:
                if not self.grid:
                    print("epoch" + str(epoch_num))

            # calc number of iteration in current epoch
            for batch_index, (A, x0, embed, label) in enumerate(self._train_loader):
                if self._gpu:
                    A, x0, embed, label = A.to(self._device), x0.to(self._device), embed.to(self._device), label.to(
                        self._device)

                # print progress
                self._model.train()

                output = self._model(A, x0, embed, test=False)  # calc output of current model on the current batch
                loss = self._loss_func(output, label.unsqueeze(dim=1).float(),
                                       weight=torch.Tensor([self._loss_weights[i].item() for i in label]).unsqueeze(
                                           dim=1).to(device=self._device)) \
                    if self._is_binary else self._loss_func(output, label,
                                                            weight=self._loss_weights.to(device=self._device))
                reg_loss = 0
                for param in self._model._qgcn_last_layer.parameters():
                    reg_loss += param.norm(norm_type)
                loss += wd * reg_loss
                # calculate loss
                loss.backward()  # back propagation
                # clip_grads(self._model, 10)

                # if (batch_index + 1) % self._batch_size == 0 or (batch_index + 1) == len_data:  # batching
                self._model.optimizer.step()  # update weights
                self._model.zero_grad()  # zero gradients

                if not self._nni:
                    if not self.grid:
                        if should_print:
                            self._print_progress(batch_index, len_data, job=TRAIN_JOB)
            # validate and print progress
            self._validate(self._train_loader, job=TRAIN_JOB)
            self._validate(self._dev_loader, job=DEV_JOB)
            self._validate(self._test_loader, job=TEST_JOB)
            if not self._nni:
                if not self.grid:
                    if should_print:
                        self._print_info(jobs=[TRAIN_JOB, DEV_JOB, TEST_JOB])

            if early_stop and epoch_num > 10 and self._print_test_loss > np.mean(self._loss_vec_train[-10:]):
                break

            dev_acc = self._print_dev_accuracy
            if dev_acc > max(self._accuracy_vec_dev):
                wait_counter = 0
            elif dev_acc < max(self._accuracy_vec_dev):
                wait_counter += 1
            if wait_counter >= wait:
                bad_counter += 1
                wait_counter = 0
                lr *= mul
                if should_print:
                    print(f"current lr is {lr}")
                for param_group in self._model.optimizer.param_groups:
                    param_group['lr'] = lr
            if bad_counter > stop_sign:
                print("early stopping")
                break

            test_acc1 = self._print_test_accuracy

        if show_plot and not self._nni:
            if not self.grid:
                self._plot_acc_dev()

        return test_acc1

    # validation function only the model and the data are important for input, the others are just for print
    def _validate(self, data_loader, job=""):
        # for calculating total loss and accuracy
        loss_count = 0
        true_labels = []
        pred = []

        self._model.eval()
        # calc number of iteration in current epoch
        len_data = len(data_loader)
        for batch_index, (A, x0, embed, label) in enumerate(data_loader):
            if self._gpu:
                A, x0, embed, label = A.to(self._device), x0.to(self._device), embed.to(self._device), label.to(
                    self._device)
            # print progress
            if not self._nni:
                if not self.grid:
                    self._print_progress(batch_index, len_data, job=VALIDATE_JOB)
            output = self._model(A, x0, embed, test=False)
            # calculate total loss
            loss_count += self._loss_func(output, label.unsqueeze(dim=1).float()) if self._is_binary else \
                self._loss_func(output, label)
            true_labels += label.tolist()
            pred += output.squeeze(dim=1).tolist()

        # update loss accuracy
        loss = float(loss_count / len(data_loader))
        self._update_loss(loss, job=job)
        self._update_accuracy_binary(pred, true_labels, job=job) if self._is_binary else \
            self._update_accuracy_multiclass(pred, true_labels, job=job)
        if self._is_binary:
            sig_pred = [sigmoid(p) for p in pred]
            self._update_auc(sig_pred, true_labels, job=job)
        return loss
